/* 
 * 
 * Copyright (c) Year(s), 2013, Luis M. S. Russo and Alexandre 
 * P. Francisco / KDBIO / INESC-ID, <qhv@kdbio.inesc-id.pt> 
 *  
 * Any published media that is related with to use of the distributed 
 * software, or derived software, must contain a reference to "Extending 
 * quick hypervolume. Luís M. S. Russo, Alexandre P. Francisco: 
 * J. Heuristics 22(3): 245-271 (2016)".
 *  
 * Permission to use, copy, modify, and/or distribute this software for 
 * any purpose with or without fee is hereby granted, provided that the 
 * above copyright notice and this permission notice appear in all 
 * copies. 
 *  
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL 
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL 
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR 
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR 
 * PERFORMANCE OF THIS SOFTWARE. 
 * 
 */

#ifndef TDATA_H
#define TDATA_H

#include "point.h"
#include "constants.h"
#include "arrayList.h"
#include "counterHash.h"

/* Some handy macros */
#define offsetof(st, m) __builtin_offsetof(st, m)
#define container_of(ptr, type, member) &(((char*)(ptr))[-offsetof(type, member)])

/* One of this exists in position of the task queue */
typedef struct taskData taskData;

/* One of this exists in position of the job queue */
typedef struct jobData jobData;

/* A box where each worker stores its tools. */
typedef struct toolBox toolBox;

struct toolBox
{
  struct iex{ /* Data for the inclusion exclusion algorithm */
    point PTS[1<<N] __attribute__ ((aligned (CLS))); /**< Initial and generated points */
    int C2P[1<<N] __attribute__ ((aligned (CLS)));   /**< Combination to position map */
  } iex;

  struct spData{ /* Data for the splitter tool */
    arrayList hoList[D+1];          /**< Array with hyperoctant lists */
    arrayList piList[D+1];          /**< Array with point index lists */
    counterHash counters;           /**< Hash with counter values */

    struct unionData{ /* Data for the union tool */
      int size;                   /**< Replaces the static size variable */
      arrayList members;              /**< List of members */
      counterHash presences;          /**< Hash of presences */
    } uD;
    struct dominantData{
      point PTS[NPOINTS];
      int localidx[NPOINTS];
    } doD;
  } spt;
};

struct jobData
{
  int n; /* The number of points in the job */
  point PS[NPOINTS]; /* The real points in the job */
  double r[NPOINTS]; /**< Store the result. */
};

/* Not necessarially the best configuration, but fairly good. */
struct taskData
{  /* Data necessary to the task */
  point z;
  point o;
  int n;
  int idx[NPOINTS];             /**< idx[i] is the point in PS (jobData) */
  double r[NPOINTS];
};

#endif /* TDATA_H */
