/* 
 * 
 * Copyright (c) Year(s), 2013, Luis M. S. Russo and Alexandre 
 * P. Francisco / KDBIO / INESC-ID, <qhv@kdbio.inesc-id.pt> 
 *  
 * Any published media that is related with to use of the distributed 
 * software, or derived software, must contain a reference to "Extending 
 * quick hypervolume. Luís M. S. Russo, Alexandre P. Francisco: 
 * J. Heuristics 22(3): 245-271 (2016)".
 *  
 * Permission to use, copy, modify, and/or distribute this software for 
 * any purpose with or without fee is hereby granted, provided that the 
 * above copyright notice and this permission notice appear in all 
 * copies. 
 *  
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL 
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL 
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR 
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR 
 * PERFORMANCE OF THIS SOFTWARE. 
 * 
 */

/**
 * @file   template.h
 * @author Luis Russo <lsr@lsr-ThinkPad-X201>
 * @date   Mon Sep  3 14:59:47 2012
 * 
 * @brief This interface abstracts away the details of the thread
 * pool.
 *
 * This particular implementation contains a job queue and a task
 * queue. Where jobs can be splitted into taks. The tasks are processed
 * by the threads. 
 *
 * Use the following steps.
 *
 * 1. Provide console and worker code:
 *    - Console posts jobs;
 *    - Workers process tasks. May involve spliting them into more tasks;
 * 2. Initialize the pool.
 * 3. Wait for the pool to finish.
 * 
 */

/* Write C-c d i to produce file information */

#ifndef TPOOL_H
#define TPOOL_H

#include "tdata.h"

/** Public functions are not aware of mutexes, and therefore do not require locks */

/** 
 * This function is used to start and finish the pool.
 * 
 * @param w The number of worker threads.
 * @param console The code to run the console. You must call pthread_exit.
 * @param run The code that is executed by a function call.
 * tid is the number of the worker.
 * @param openBox Set up the worker's tool box. So he can start working. 
 * Assume that the tools are autocleaning between jobs.
 * @param priority The priority to assign a given task.
 * @param jobCopy Copy the job data to the first argument. Used when copying to job queue.
 * @param taskCopy Copy the task data to the first argument. Used when copying to task queue.
 * @param merge Merge the taskData in src into the result in dst.
 * @param publish the computed result. Depends on the structure of the jobData.
 */
void launch(
            int w,
            void* (*console)(void *threadid),
            void  (*run)(toolBox* b, taskData* d, const jobData* jd), 
            void  (*openBox)(toolBox* b),
            void  (*closeBox)(toolBox* b),
            int   (*priority)(taskData* d),
            void  (*jobCopy)(jobData* des, jobData* src),
            void  (*taskCopy)(taskData* dst, taskData* src),
            void  (*merge)(jobData* dst, taskData* src),
            void  (*publish)(int jid, jobData* jd)
            );
/** 
 *
 * Executes run on d. If it does not access the queue, or submit
 * fails, it stores the result back in d, incrementally. If the work
 * goes to the queue it will eventually be merged back in the
 * respective jobData.
 * 
 * @param i The task to compute.
 * @param p A boolean indication whether to paralelize the computation.
 */
void prun(taskData* d, int p);

/** 
 *
 * Starts a new job, by uploading the first task.
 *
 * @param jd Constant job data to upload to the queue.
 * @param i The task to upload to the queue.
 */
void postJob(jobData* jd, taskData* d);

/* The following functions can be used to upload jobs and tasks. */

/*** typedefs(not structures) and defined constants *******/
	
/*** enums ************************************************/
	
/*** structures declarations (only if really necessary) ***/
	
/*** declarations of public functions (first inlines) *****/

#endif /* TPOOL_H */
